/*****************************************************************************
 *
 *  gkv5500 - gtk2 frontend of kv5500 utility.
 *
 * Copyright (c) 2006-2007 Hyo geol, Lee
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 *   3. The name of the author may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************/
#include <gtk/gtk.h>

#include <libintl.h>
#include <locale.h>

#include <errno.h>
#include <assert.h>

#include <fstream>
#include <sstream>

#include "gtkhg.hpp"

#include "gkv5500.hpp"
#include "kv5500.hpp"
#include "kv5500_sms.hpp"
#include "kv5500_phonebook.hpp"
#include "kv5500_photo.hpp"
#include "kv5500_video.hpp"
#include "gkv5500_file.hpp"

using namespace KV5500;
using namespace HG;

typedef GetFileScanner<PhotoInfo, CopyFilenameGenerator, GtkDisplay> GPhotoGetFileScanner;

typedef GetFileScanner<VideoInfo, CopyFilenameGenerator, GtkDisplay> GVideoGetFileScanner;

GtkWindow* KV5500::gMainWindow = 0;

void SensitiveMenu::On()
{
    gtk_widget_set_sensitive(&save_, TRUE);
    gtk_widget_set_sensitive(&saveas_, TRUE);
    gtk_widget_set_sensitive(&saveTool_, TRUE);
    gtk_widget_set_sensitive(&saveasTool_, TRUE);
    gtk_widget_set_sensitive(&clearTool_, TRUE);
}

void SensitiveMenu::Off()
{
    gtk_widget_set_sensitive(&save_, FALSE);
    gtk_widget_set_sensitive(&saveas_, FALSE);
    gtk_widget_set_sensitive(&saveTool_, FALSE);
    gtk_widget_set_sensitive(&saveasTool_, FALSE);
    gtk_widget_set_sensitive(&clearTool_, FALSE);
}

std::ostream& KV5500::PrintSeparator(std::ostream& os, unsigned int current, unsigned int total)
{
    os << "---------------------------------";

    os.width(3);
    os.fill('0');
    os << current;

    os << '/';

    os.width(3);
    os.fill('0');
    os << total;

    os << "------\n";

    return os;
}

GtkWidget* KV5500::MakeSaveAsDialog(const char* const defaultFilename)
{
    assert(defaultFilename);

    GtkWidget* const fileDialog = gtk_file_chooser_dialog_new(gettext("Save as ..."),
                                                              gMainWindow,
                                                              GTK_FILE_CHOOSER_ACTION_SAVE,
                                                              GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                                              GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
                                                              NULL);
    assert(fileDialog);
    gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(fileDialog), TRUE);

    gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(fileDialog), defaultFilename);

    return fileDialog;
}

const std::string KV5500::GetFilesizeStr(unsigned int size)
{
    std::ostringstream ss;

    ss << size / 1024.0 / 1024.0 << gettext(" MBytes");

    return ss.str();
}

const int KV5500::GetFileNumber(GtkNotebook& notebook, gint page)
{
    GtkTreeView* const treeView = GetTreeView(notebook, page);

    const GtkTreePathHelper path(treeView);
    if(path.IsExist()){

        const GCharPtr pathString(path.GetString());
        
        const int rtn = strtol(pathString.Get(), 0, 10);
        assert(rtn >= 0);

        return (rtn == 0 && (errno == EINVAL || errno == ERANGE)) ? -1 : rtn + 1;
    }

    return -1;
}

CharArrayPtr KV5500::DateStrNow()
{
    size_t dateStrSize = gDateStrSize;

    time_t now;
    time(&now);

    const tm* const ptm = localtime(&now);
    assert(ptm);
    CharArrayPtr dateStr;

    do {
        dateStr = CharArrayPtr(new char[dateStrSize]);

        const size_t write = strftime(dateStr.Get(), dateStrSize, "%F-%T", ptm);

        if(write == 0) {
            dateStrSize = static_cast<size_t>(dateStrSize * gDateStrGrowFactor);
        }

    } while(write == 0);

    return dateStr;
}

const std::string KV5500::GetDateFilename(const char* prefix, const char* postfix)
{
    assert(prefix && postfix);

    std::ostringstream ss;

    ss << prefix;

    const CharArrayPtr now = DateStrNow();
    assert(now.Get());
    ss << now.Get();

    ss << postfix;

    return ss.str();
}

GtkTreeView* KV5500::GetTreeView(GtkNotebook& notebook, gint page)
{
    GtkScrolledWindow* const scrolledWindow = GTK_SCROLLED_WINDOW(gtk_notebook_get_nth_page(&notebook, page));
    assert(scrolledWindow);

    GtkTreeView* const treeView = GTK_TREE_VIEW(gtk_bin_get_child(GTK_BIN(scrolledWindow)));
    assert(treeView);

    return treeView;
}

GtkTreeModel* KV5500::GetTreeModel(GtkNotebook& notebook, gint page)
{
    GtkTreeView* const treeView = GetTreeView(notebook, page);

    GtkTreeModel* const treeModel = gtk_tree_view_get_model(treeView);
    assert(treeModel);

    return treeModel;
}

static void KV5500::OnDestory(GtkWidget*, gpointer)
{
    gtk_main_quit();
}

void KV5500::RegistSMS(const std::vector<SMS>& sms, GtkListStore& listStore)
{
    assert(!sms.empty());

    GtkTreeIter iter;

    gtk_list_store_clear(&listStore);

    const std::vector<SMS>::const_iterator end = sms.end();
    for(std::vector<SMS>::const_iterator i = sms.begin();
        i != end;
        ++i) {

        gsize in = i->GetMessage().size();
        gsize out = 0;

        const GCharPtr utf8(g_convert(i->GetMessage().c_str(),
                                      -1,
                                      gUtf8CodesetStr,
                                      gKV5500CodesetStr,
                                      &in,
                                      &out,
                                      NULL));
        
        if(utf8.Get() != NULL) {

            gtk_list_store_append(&listStore, &iter);

            gtk_list_store_set(&listStore, &iter,
                               SMS_COL_MESSAGE,
                               utf8.Get(),
                               SMS_COL_NUMBER,
                               i->GetNumber().c_str(),
                               SMS_COL_DATE,
                               i->GetDate().c_str(),
                               -1);
        }
    }
}

void KV5500::RegistPhonebook(const std::vector<PhonebookData>& phonebook, GtkTreeStore& treeStore)
{
    assert(!phonebook.empty());

    GtkTreeIter parent;
    GtkTreeIter iter;

    gtk_tree_store_clear(&treeStore);

    const std::vector<PhonebookData>::const_iterator end = phonebook.end();
    for(std::vector<PhonebookData>::const_iterator i = phonebook.begin();
        i != end;
        ++i) {

        gsize in = i->GetName().size();
        gsize out = 0;

        const GCharPtr utf8(g_convert(i->GetName().c_str(),
                                      -1,
                                      gUtf8CodesetStr,
                                      gKV5500CodesetStr,
                                      &in,
                                      &out,
                                      NULL));

        if(utf8.Get() != NULL) {

            gtk_tree_store_append(&treeStore, &iter, NULL);
            parent = iter;
            gtk_tree_store_set(&treeStore, &iter,
                               PHONE_COL_NAME,
                               utf8.Get(),
                               -1);

            const std::vector<std::string>& numbers = i->GetNumber();
            assert(!numbers.empty());
            const std::vector<std::string>::const_iterator endNumbers = numbers.end();
            for(std::vector<std::string>::const_iterator j = numbers.begin();
                j != endNumbers;
                ++j) {

                gtk_tree_store_append(&treeStore, &iter, &parent);
                gtk_tree_store_set(&treeStore, &iter,
                                   PHONE_COL_NAME,
                                   j->c_str(),
                                   -1);
            }
        }
    }
}

void KV5500::RegistFileInfo(const std::vector<FileInfo>& fileInfo, GtkListStore& listStore)
{
    assert(!fileInfo.empty());

    gtk_list_store_clear(&listStore);

    GtkTreeIter iter;

    const std::vector<FileInfo>::const_iterator end = fileInfo.end();
    for(std::vector<FileInfo>::const_iterator i = fileInfo.begin();
        i != end;
        ++i) {

        gsize in = i->GetName().size();
        gsize out = 0;

        const GCharPtr utf8(g_convert(i->GetName().c_str(),
                                      -1,
                                      gUtf8CodesetStr,
                                      gKV5500CodesetStr,
                                      &in,
                                      &out,
                                      NULL));

        if(utf8.Get() != NULL) {

            const std::string fileSize = GetFilesizeStr(i->GetSize());
            gtk_list_store_append(&listStore, &iter);
            gtk_list_store_set(&listStore, &iter,
                               FILE_COL_NAME,
                               utf8.Get(),
                               FILE_COL_SIZE,
                               fileSize.c_str(),
                               -1);
        }
    }
}

int KV5500::SMSScan(Usb& usb, GtkNotebook& notebook)
{
    SMSScanner smsScanner;
    if(usb.Scan(smsScanner)) {

        std::vector<SMS> thisSMS;
        smsScanner.GetSMS(thisSMS);

        if(!thisSMS.empty()) {

            GtkListStore* const smsListStore = GTK_LIST_STORE(GetTreeModel(notebook, SMS_PAGE));

            GdkLock lock;
            RegistSMS(thisSMS, *smsListStore);
        }

        return thisSMS.size();
    }

    return -1;
}

void* KV5500::GetSMSThread(void* arg)
{
    assert(arg);

    ConnectThreadArg* const tArg = static_cast<ConnectThreadArg*>(arg);

    assert(tArg->dialog_ && tArg->msgLabel_ && tArg->closeBtn_);

    if(!tArg->usb_.Find()) {

        GdkLock lock;

        gtk_label_set_text(tArg->msgLabel_,
                           gettext("Error : usb find functions."));

        gtk_widget_set_sensitive(tArg->closeBtn_, TRUE);

        return 0;
    }

    std::ostringstream ss;

    try {

        const int recvSize = SMSScan(tArg->usb_, tArg->notebook_);
        if(recvSize > 0) {

            ss << gettext("Done, total ");
            ss << recvSize;
            ss << gettext(" messages received.");

            GdkLock lock;
            tArg->sMenu_.On();

        } else if(recvSize == 0) {

            ss << gettext("No message.");

        } else {

            ss << gettext("Error : Cannot get SMS.");
        }

    } catch(std::exception& e) {

        ss << gettext("Error : ");
        ss << e.what();
    }

    GdkLock lock;

    gtk_label_set_text(tArg->msgLabel_, ss.str().c_str());

    gtk_widget_set_sensitive(tArg->closeBtn_, TRUE);

    return 0;
}

int KV5500::PhonebookScan(Usb& usb, GtkNotebook& notebook)
{
    PhonebookScanner pbScanner;

    if(usb.Scan(pbScanner)) {

        std::vector<PhonebookData> thisPhonebook;
        pbScanner.GetPhonebook(thisPhonebook);

        if(!thisPhonebook.empty()) {

            GtkTreeStore* const phonebookTreeStore = GTK_TREE_STORE(GetTreeModel(notebook, PHONEBOOK_PAGE));

            GdkLock lock;

            RegistPhonebook(thisPhonebook, *phonebookTreeStore);
        }
        
        return thisPhonebook.size();
    }

    return -1;
}

void* KV5500::GetPhonebookThread(void* arg)
{
    assert(arg);

    ConnectThreadArg* const tArg = static_cast<ConnectThreadArg*>(arg);

    assert(tArg->dialog_ && tArg->msgLabel_ && tArg->closeBtn_);

    if(!tArg->usb_.Find()) {

        GdkLock lock;

        gtk_label_set_text(tArg->msgLabel_,
                           gettext("Error : usb find functions."));

        gtk_widget_set_sensitive(tArg->closeBtn_, TRUE);

        return 0;
    }

    std::ostringstream ss;

    try {

        const int recvSize = PhonebookScan(tArg->usb_, tArg->notebook_);
        if(recvSize > 0) {

            ss << gettext("Done, total ");
            ss << recvSize;
            ss << gettext(" person of phonebook received.");

            tArg->sMenu_.On();

        } else if(recvSize == 0) {

            ss << gettext("No person.");

        } else {

            ss << gettext("Error : Cannot get phonebook.");
        }

    } catch(std::exception& e) {

        ss << gettext("Error : ");
        ss << e.what();
    }

    GdkLock lock;

    gtk_label_set_text(tArg->msgLabel_, ss.str().c_str());

    gtk_widget_set_sensitive(tArg->closeBtn_, TRUE);

    return 0;
}

int KV5500::PhotoInfoScan(Usb& usb, GtkNotebook& notebook)
{
    PhotoInfoScanner piScanner;

    if(usb.Scan(piScanner)) {

        std::vector<FileInfo> thisPhotoList;
        piScanner.GetInfo(thisPhotoList);

        if(!thisPhotoList.empty()) {

            GtkListStore* const photoListStore = GTK_LIST_STORE(GetTreeModel(notebook, PHOTO_PAGE));

            GdkLock lock;

            RegistFileInfo(thisPhotoList, *photoListStore);
        }

        return thisPhotoList.size();
    }

    return -1;
}

void* KV5500::GetPhotoListThread(void* arg)
{
    assert(arg);

    ConnectThreadArg* const tArg = static_cast<ConnectThreadArg*>(arg);

    assert(tArg->dialog_ && tArg->msgLabel_ && tArg->closeBtn_);

    if(!tArg->usb_.Find()) {

        GdkLock lock;

        gtk_label_set_text(tArg->msgLabel_,
                           gettext("Error : usb find functions."));

        gtk_widget_set_sensitive(tArg->closeBtn_, TRUE);

        return 0;
    }

    std::ostringstream ss;

    try {

        const int recvSize = PhotoInfoScan(tArg->usb_, tArg->notebook_);
        if(recvSize > 0) {

            ss << gettext("Done, total ");
            ss << recvSize;
            ss << gettext(" photo info received.");

            tArg->sMenu_.On();

        } else if(recvSize == 0) {

            ss << gettext("No photo.");

        } else {

            ss << gettext("Error : Cannot get photo list.");
        }

    } catch(std::exception& e) {

        ss << gettext("Error : ");
        ss << e.what();
    }

    GdkLock lock;

    gtk_label_set_text(tArg->msgLabel_, ss.str().c_str());

    gtk_widget_set_sensitive(tArg->closeBtn_, TRUE);

    return 0;
}

int KV5500::VideoInfoScan(Usb& usb, GtkNotebook& notebook)
{
    VideoInfoScanner viScanner;

    if(usb.Scan(viScanner)) {

        std::vector<FileInfo> thisVideoList;
        viScanner.GetInfo(thisVideoList);

        if(!thisVideoList.empty()) {

            GtkListStore* const videoListStore = GTK_LIST_STORE(GetTreeModel(notebook, VIDEO_PAGE));

            GdkLock lock;

            RegistFileInfo(thisVideoList, *videoListStore);
        }

        return thisVideoList.size();
    }

    return -1;
}

void* KV5500::GetVideoListThread(void* arg)
{
    assert(arg);

    ConnectThreadArg* const tArg = static_cast<ConnectThreadArg*>(arg);

    assert(tArg->dialog_ && tArg->msgLabel_ && tArg->closeBtn_);

    if(!tArg->usb_.Find()) {

        GdkLock lock;

        gtk_label_set_text(tArg->msgLabel_,
                           gettext("Error : usb find functions."));

        gtk_widget_set_sensitive(tArg->closeBtn_, TRUE);

        return 0;
    }

    std::ostringstream ss;

    try {

        const int recvSize = VideoInfoScan(tArg->usb_, tArg->notebook_);
        if(recvSize > 0) {

            ss << gettext("Done, total ");
            ss << recvSize;
            ss << gettext(" video info received.");

            tArg->sMenu_.On();

        } else if(recvSize == 0) {

            ss << gettext("No video.");

        } else {

            ss << gettext("Error : Cannot get video list.");
        }

    } catch(std::exception& e) {

        ss << gettext("Error : ");
        ss << e.what();
    }

    GdkLock lock;

    gtk_label_set_text(tArg->msgLabel_, ss.str().c_str());

    gtk_widget_set_sensitive(tArg->closeBtn_, TRUE);

    return 0;
}

void KV5500::MakeConnectDialog(ConnectThreadArg& tArg, const gchar* title, const gchar* defaultLabel)
{
    assert(title && defaultLabel);

    GtkWidget* const dialog = gtk_dialog_new();
    assert(dialog);
    tArg.dialog_ = GTK_DIALOG(dialog);

    gtk_window_set_transient_for(GTK_WINDOW(dialog), gMainWindow);
    gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
    gtk_window_set_title(GTK_WINDOW(dialog), title);

    GtkWidget* const connectLabel = gtk_label_new(defaultLabel);
    assert(connectLabel);
    gtk_box_pack_start(GTK_BOX( GTK_DIALOG(dialog) ->vbox ), connectLabel, 1, 1, 0);
    tArg.msgLabel_ = GTK_LABEL(connectLabel);
    GtkWidget* const closeBtn = gtk_dialog_add_button(GTK_DIALOG(dialog),
                                                "gtk-close",
                                                0);
    assert(closeBtn);
    tArg.closeBtn_ = closeBtn;
    gtk_widget_set_sensitive(closeBtn, FALSE);

    gtk_window_set_default_size(GTK_WINDOW(dialog), 261, 161);
}

void KV5500::ConnectSMS(ConnectThreadArg& tArg)
{
    MakeConnectDialog(tArg, gettext("Get SMS ..."), gettext("Connecting to get SMS ..."));

    GThread* const thread = g_thread_create(GetSMSThread, &tArg, TRUE, 0);

    if(thread) {
        gtk_widget_show_all(GTK_WIDGET(tArg.dialog_));
        gtk_dialog_run(tArg.dialog_);
        g_thread_join(thread);
    }

    gtk_widget_destroy(GTK_WIDGET(tArg.dialog_));
}

void KV5500::ConnectPhonebook(ConnectThreadArg& tArg)
{
    MakeConnectDialog(tArg, gettext("Get phonebook ..."), gettext("Connecting to get phonebook ..."));

    GThread* const thread = g_thread_create(GetPhonebookThread, &tArg, TRUE, 0);

    if(thread) {
        gtk_widget_show_all(GTK_WIDGET(tArg.dialog_));
        gtk_dialog_run(tArg.dialog_);
        g_thread_join(thread);
    }

    gtk_widget_destroy(GTK_WIDGET(tArg.dialog_));
}

void KV5500::ConnectPhoto(ConnectThreadArg& tArg)
{
    MakeConnectDialog(tArg, gettext("Get photo list ... "), gettext("Connecting to get photo list ..."));

    GThread* const thread = g_thread_create(GetPhotoListThread, &tArg, TRUE, 0);

    if(thread) {
        gtk_widget_show_all(GTK_WIDGET(tArg.dialog_));
        gtk_dialog_run(tArg.dialog_);
        g_thread_join(thread);
    }

    gtk_widget_destroy(GTK_WIDGET(tArg.dialog_));
}

void KV5500::ConnectVideo(ConnectThreadArg& tArg)
{
    MakeConnectDialog(tArg, gettext("Get video list ... "), gettext("Connecting to get video list ..."));

    GThread* const thread = g_thread_create(GetVideoListThread, &tArg, TRUE, 0);

    if(thread) {
        gtk_widget_show_all(GTK_WIDGET(tArg.dialog_));
        gtk_dialog_run(tArg.dialog_);
        g_thread_join(thread);
    }

    gtk_widget_destroy(GTK_WIDGET(tArg.dialog_));
}

static void KV5500::OnActivateConnect(GtkMenuItem*, gpointer data)
{
    assert(data);

    ConnectThreadArg* const tArg = static_cast<ConnectThreadArg*>(data);

    switch(gtk_notebook_get_current_page(&tArg->notebook_)) {
    case SMS_PAGE:

        ConnectSMS(*tArg);

        break;

    case PHONEBOOK_PAGE:

        ConnectPhonebook(*tArg);

        break;

    case PHOTO_PAGE:

        ConnectPhoto(*tArg);

        break;

    case VIDEO_PAGE:

        ConnectVideo(*tArg);

        break;

    default:
        break;
    }
}

void KV5500::ShowSavedDialog(const char* filename)
{
    assert(filename);

    GtkWidget* const okDialog = gtk_message_dialog_new(gMainWindow,
                                                       GTK_DIALOG_MODAL,
                                                       GTK_MESSAGE_INFO,
                                                       GTK_BUTTONS_CLOSE,
                                                       gettext("Saved to file : %s"),
                                                       filename);
    assert(okDialog);

    gtk_dialog_run(GTK_DIALOG(okDialog));

    gtk_widget_destroy(okDialog);
}

void KV5500::ShowSaveErrDialog(const char* filename)
{
    assert(filename);

    GtkWidget* const errDialog = gtk_message_dialog_new(gMainWindow,
                                                        GTK_DIALOG_MODAL,
                                                        GTK_MESSAGE_ERROR,
                                                        GTK_BUTTONS_CLOSE,
                                                        gettext("Cannot open file : %s"),
                                                        filename);
    assert(errDialog);

    gtk_dialog_run(GTK_DIALOG(errDialog));

    gtk_widget_destroy(errDialog);
}

GtkWidget* KV5500::MakeTransferDialog(GtkLabel*& label, GtkWidget*& progressBar, GtkWidget*& closeBtn)
{
    GtkWidget* const dialog = gtk_dialog_new();
    assert(dialog);

    gtk_window_set_transient_for(GTK_WINDOW(dialog), gMainWindow);
    gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
    gtk_window_set_title(GTK_WINDOW(dialog), gettext("Transfering ..."));
    label = GTK_LABEL(gtk_label_new(gettext("Connecting to transfer file ...")));
    assert(label);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), GTK_WIDGET(label), 1, 1, 0);

    progressBar = gtk_progress_bar_new();
    assert(progressBar);
    gtk_widget_set_sensitive(progressBar, FALSE);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), progressBar, 0, 1, 0);

    closeBtn = gtk_dialog_add_button(GTK_DIALOG(dialog),
                                     "gtk-close",
                                     0);
    assert(closeBtn);

    gtk_widget_set_sensitive(closeBtn, FALSE);
    gtk_window_set_default_size(GTK_WINDOW(dialog),
                                gTransferDialogWidth,
                                gTransferDialogHeight);

    return dialog;
}

/// @todo Permission check of output file, directory.
void KV5500::SaveSMS(GtkNotebook& notebook, const char* filename)
{
    assert(filename);

    std::ofstream os(filename);
    if(os) {

        GtkTreeModel* const model = GetTreeModel(notebook, SMS_PAGE);

        const gint total = gtk_tree_model_iter_n_children(model, NULL);

        GtkTreeIter iter;
        unsigned int count = 1;
        bool valid = gtk_tree_model_get_iter_first(model, &iter);

        while(valid) {

            gchar* message = 0;
            gchar* number = 0;
            gchar* date = 0;

            gtk_tree_model_get(model, &iter,
                               SMS_COL_MESSAGE, &message,
                               SMS_COL_NUMBER, &number,
                               SMS_COL_DATE, &date,
                               -1);

            assert(message && number && date);

            GCharPtr messageFree(message);
            GCharPtr numberFree(number);
            GCharPtr dateFree(date);

            PrintSeparator(os, count, total);

            os << " Number  : " << number << '\n';
            os << " Date    : " << date << '\n';
            os << " Message : " << message << '\n';

            valid = gtk_tree_model_iter_next(model, &iter);
            ++count;
        }

        ShowSavedDialog(filename);

    } else {

        ShowSaveErrDialog(filename);
    }
}

/// @todo Permission check of output file, directory.
void KV5500::SavePhonebook(GtkNotebook& notebook, const char* filename)
{
    assert(filename);

    std::ofstream os(filename);
    if(os) {

        GtkTreeModel* const model = GetTreeModel(notebook, PHONEBOOK_PAGE);

        const gint total = gtk_tree_model_iter_n_children(model, NULL);

        GtkTreeIter iter;
        GtkTreeIter child;
        unsigned int count = 1;
        bool valid = gtk_tree_model_get_iter_first(model, &iter);

        while(valid) {

            gchar* name = 0;
            gtk_tree_model_get(model, &iter,
                               PHONE_COL_NAME, &name,
                               -1);
            assert(name);
            GCharPtr nameFree(name);

            PrintSeparator(os, count, total);

            os << " Name     : " << name << '\n';

            bool validChild = gtk_tree_model_iter_children(model, &child, &iter);
            unsigned int numCount = 1;

            while(validChild) {

                gchar* number = 0;
                gtk_tree_model_get(model, &child,
                                   PHONE_COL_NAME, &number,
                                   -1);
                assert(number);
                GCharPtr numberFree(number);

                os << " Number " << numCount << " : " << number << '\n';

                validChild = gtk_tree_model_iter_next(model, &child);
                ++numCount;
            }

            valid = gtk_tree_model_iter_next(model, &iter);
            ++count;
        }

        ShowSavedDialog(filename);

    } else {

        ShowSaveErrDialog(filename);
    }
}

void* KV5500::TransferPhotoThread(void* arg)
{
    assert(arg);

    TransferThreadArg* const tArg = static_cast<TransferThreadArg*>(arg);

    if(!tArg->usb_.Find()) {

        GdkLock lock;

        gtk_label_set_text(&tArg->msgLabel_,
                           gettext("Error : usb find functions."));
        gtk_widget_set_sensitive(&tArg->closeBtn_, TRUE);

        return 0;
    }

    std::ostringstream ss;

    try {

        CopyFilenameGenerator ng(tArg->filename_);
        GtkDisplay gd(GTK_WIDGET(&tArg->msgLabel_), &tArg->progressBar_);
        GPhotoGetFileScanner pgScanner(tArg->fileNum_, ng, gd);
        if(!tArg->usb_.Scan(pgScanner))
            ss << gettext("Error : Cannot transfer photo file.");

    } catch(std::exception& e) {

        ss << gettext("Error : ");
        ss << e.what();
    }

    GdkLock lock;

    if(!ss.str().empty())
        gtk_label_set_text(&tArg->msgLabel_, ss.str().c_str());

    gtk_widget_set_sensitive(&tArg->closeBtn_, TRUE);

    return 0;
}

void KV5500::SavePhoto(Usb& usb, unsigned int fileNum, const std::string& filename)
{
    GtkLabel* label;
    GtkWidget* progressBar;
    GtkWidget* closeBtn;
    GtkWidget* const dialog = MakeTransferDialog(label, progressBar, closeBtn);

    TransferThreadArg transArg(usb, *label, *progressBar, *closeBtn, fileNum, filename);

    GThread* const thread = g_thread_create(TransferPhotoThread, &transArg, TRUE, 0);

    if(thread) {
        gtk_widget_show_all(dialog);
        gtk_dialog_run(GTK_DIALOG(dialog));
        g_thread_join(thread);
    }

    gtk_widget_destroy(dialog);
}

void* KV5500::TransferVideoThread(void* arg)
{
    assert(arg);

    TransferThreadArg* const tArg = static_cast<TransferThreadArg*>(arg);

    if(!tArg->usb_.Find()) {

        GdkLock lock;

        gtk_label_set_text(&tArg->msgLabel_,
                           gettext("Error : usb find functions."));
        gtk_widget_set_sensitive(&tArg->closeBtn_, TRUE);

        return 0;
    }

    std::ostringstream ss;

    try {

        CopyFilenameGenerator ng(tArg->filename_);
        GtkDisplay gd(GTK_WIDGET(&tArg->msgLabel_), &tArg->progressBar_);
        GVideoGetFileScanner vgScanner(tArg->fileNum_, ng, gd);
        if(!tArg->usb_.Scan(vgScanner))
            ss << gettext("Error : Cannot transfer photo file.");

    } catch(std::exception& e) {

        ss << gettext("Error : ");
        ss << e.what();
    }

    GdkLock lock;

    if(!ss.str().empty())
        gtk_label_set_text(&tArg->msgLabel_, ss.str().c_str());

    gtk_widget_set_sensitive(&tArg->closeBtn_, TRUE);

    return 0;
}

void KV5500::SaveVideo(Usb& usb, unsigned int fileNum, const std::string& filename)
{
    GtkLabel* label;
    GtkWidget* progressBar;
    GtkWidget* closeBtn;
    GtkWidget* const dialog = MakeTransferDialog(label, progressBar, closeBtn);

    TransferThreadArg transArg(usb, *label, *progressBar, *closeBtn, fileNum, filename);

    GThread* const thread = g_thread_create(TransferVideoThread, &transArg, TRUE, 0);

    if(thread) {
        gtk_widget_show_all(dialog);
        gtk_dialog_run(GTK_DIALOG(dialog));
        g_thread_join(thread);
    }

    gtk_widget_destroy(dialog);
}

static void KV5500::OnActivateSave(GtkMenuItem*, gpointer data)
{
    assert(data);

    SaveWidgetArg* const sArg = static_cast<SaveWidgetArg*>(data);

    switch(gtk_notebook_get_current_page(&sArg->notebook_)) {
    case SMS_PAGE:

        {
            const std::string filename = GetDateFilename("gkv5500_sms_", "_UTF-8.txt");
            SaveSMS(sArg->notebook_, filename.c_str());
        }

        break;

    case PHONEBOOK_PAGE:

        {
            const std::string filename = GetDateFilename("gkv5500_phonebook_", "_UTF-8.txt");
            SavePhonebook(sArg->notebook_, filename.c_str());
        }

        break;

    case PHOTO_PAGE:

        {
            const int fileNum = GetFileNumber(sArg->notebook_, PHOTO_PAGE);
            if(fileNum > 0) {
                std::ostringstream ss;
                ss << "gkv5500_photo_";
                ss.width(3);
                ss.fill('0');
                ss << fileNum;
                ss << '_';

                const std::string filename = GetDateFilename(ss.str().c_str(), ".jpg");
                SavePhoto(sArg->usb_, fileNum, filename);
            }
        }

        break;

    case VIDEO_PAGE:

        {
            const int fileNum = GetFileNumber(sArg->notebook_, VIDEO_PAGE);
            if(fileNum > 0) {
                std::ostringstream ss;
                ss << "gkv5500_video_";
                ss.width(3);
                ss.fill('0');
                ss << fileNum;
                ss << '_';

                const std::string filename = GetDateFilename(ss.str().c_str(), ".avi");
                SaveVideo(sArg->usb_, fileNum, filename);
            }
        }

        break;

    default:
        break;
    }
}

static void KV5500::OnActivateSaveAs(GtkMenuItem*, gpointer data)
{
    assert(data);

    SaveWidgetArg* const sArg = static_cast<SaveWidgetArg*>(data);

    switch(gtk_notebook_get_current_page(&sArg->notebook_)) {
    case SMS_PAGE:

        {
            const std::string defaultFilename = GetDateFilename("gkv5500_sms_", "_UTF-8.txt");

            GtkWidget* const fileDialog = MakeSaveAsDialog(defaultFilename.c_str());
            assert(fileDialog);

            const gint res = gtk_dialog_run(GTK_DIALOG(fileDialog));

            const DestPtr<const char, GFreeDestructor> filename(gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fileDialog)));

            gtk_widget_destroy(fileDialog);

            if(res == GTK_RESPONSE_ACCEPT) {

                SaveSMS(sArg->notebook_, filename.Get());
            }
        }

        break;

    case PHONEBOOK_PAGE:

        {
            const std::string defaultFilename = GetDateFilename("gkv5500_phonebook_", "_UTF-8.txt");

            GtkWidget* const fileDialog = MakeSaveAsDialog(defaultFilename.c_str());
            assert(fileDialog);
            
            const gint res = gtk_dialog_run(GTK_DIALOG(fileDialog));
            const DestPtr<const char, GFreeDestructor> filename(gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fileDialog)));

            gtk_widget_destroy(fileDialog);

            if(res == GTK_RESPONSE_ACCEPT) {

                SavePhonebook(sArg->notebook_, filename.Get());
            }
        }

        break;

    case PHOTO_PAGE:

        {
            const int fileNum = GetFileNumber(sArg->notebook_, PHOTO_PAGE);
            if(fileNum > 0) {
                std::ostringstream ss;
                ss << "gkv5500_photo_";
                ss.width(3);
                ss.fill('0');
                ss << fileNum;
                ss << '_';

                const std::string defaultFilename = GetDateFilename(ss.str().c_str(), ".jpg");

                GtkWidget* const fileDialog = MakeSaveAsDialog(defaultFilename.c_str());
                assert(fileDialog);
            
                const gint res = gtk_dialog_run(GTK_DIALOG(fileDialog));
                const DestPtr<const char, GFreeDestructor> filename(gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fileDialog)));

                gtk_widget_destroy(fileDialog);

                if(res == GTK_RESPONSE_ACCEPT) {

                    SavePhoto(sArg->usb_, fileNum, filename.Get());
                }
            }
        }

        break;

    case VIDEO_PAGE:

        {
            const int fileNum = GetFileNumber(sArg->notebook_, VIDEO_PAGE);
            if(fileNum > 0) {
                std::ostringstream ss;
                ss << "gkv5500_video_";
                ss.width(3);
                ss.fill('0');
                ss << fileNum;
                ss << '_';

                const std::string defaultFilename = GetDateFilename(ss.str().c_str(), ".avi");

                GtkWidget* const fileDialog = MakeSaveAsDialog(defaultFilename.c_str());
                assert(fileDialog);
            
                const gint res = gtk_dialog_run(GTK_DIALOG(fileDialog));
                const DestPtr<const char, GFreeDestructor> filename(gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fileDialog)));

                gtk_widget_destroy(fileDialog);

                if(res == GTK_RESPONSE_ACCEPT) {

                    SaveVideo(sArg->usb_, fileNum, filename.Get());
                }
            }
        }

        break;

    default:
        break;
    }
}

static void KV5500::OnClickedClear(GtkMenuItem*, gpointer data)
{
    assert(data);

    ClearWidgetArg* const cArg = static_cast<ClearWidgetArg*>(data);

    const gint currentPage = gtk_notebook_get_current_page(&cArg->notebook_);
    assert(currentPage >= 0);

    GtkWidget* const treeView = gtk_bin_get_child(GTK_BIN(gtk_notebook_get_nth_page(&cArg->notebook_, currentPage)));
    assert(treeView);

    if(currentPage == PHONEBOOK_PAGE) {

        GtkTreeStore* const treeStore = GTK_TREE_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(treeView)));
        assert(treeStore);
        gtk_tree_store_clear(treeStore);

    } else {

        GtkListStore* const listStore = GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(treeView)));
        assert(listStore);
        gtk_list_store_clear(listStore);
    }

    cArg->sMenu_.Off();
}

static void KV5500::OnActivateAbout(GtkMenuItem*, gpointer data)
{
    assert(data);

    const gchar* authors[] = { gettext("Hyo geol, Lee <hyogeollee@gmail.com>"), NULL };

    GtkWidget* const dialog = gtk_about_dialog_new();
    assert(dialog);

    gtk_window_set_transient_for(GTK_WINDOW(dialog), GTK_WINDOW(data));

    gtk_about_dialog_set_version(GTK_ABOUT_DIALOG(dialog), VERSION);
    gtk_about_dialog_set_copyright(GTK_ABOUT_DIALOG(dialog),
                                   gettext("BSD license"));
    gtk_about_dialog_set_website(GTK_ABOUT_DIALOG(dialog),
                                 "http://kldp.net/projects/kv5500");
    gtk_about_dialog_set_authors(GTK_ABOUT_DIALOG(dialog),
                                 authors);

    GdkPixbuf* const icon = gtk_window_get_icon(gMainWindow);
    if(icon)
        gtk_about_dialog_set_logo(GTK_ABOUT_DIALOG(dialog), icon);

    gtk_dialog_run(GTK_DIALOG(dialog));

    gtk_widget_destroy(GTK_WIDGET(dialog));
}

static void KV5500::OnSwitchPage(GtkNotebook* notebook,
                                 GtkNotebookPage*,
                                 guint pageNum,
                                 gpointer data)
{
    assert(notebook && data);

    GtkWidget* const treeView = gtk_bin_get_child(GTK_BIN(gtk_notebook_get_nth_page(notebook, pageNum)));
    assert(treeView);

    GtkTreeModel* const model = gtk_tree_view_get_model(GTK_TREE_VIEW(treeView));
    assert(model);
    GtkTreeIter iter;
    static_cast<SensitiveMenu*>(data)->Enable(gtk_tree_model_get_iter_first(model, &iter));
}

static void KV5500::OnSMSRowActivated(GtkTreeView* treeview, GtkTreePath* path, GtkTreeViewColumn* column, gpointer data)
{
    assert(treeview && path && column && data);

    GtkTreeModel* const model = gtk_tree_view_get_model(treeview);
    assert(model);
    GtkTreeIter iter;
    if(gtk_tree_model_get_iter(model, &iter, path)) {

        gchar* message = 0;
        gchar* number = 0;
        gchar* date = 0;

        gtk_tree_model_get(model, &iter,
                           SMS_COL_MESSAGE, &message,
                           SMS_COL_NUMBER, &number,
                           SMS_COL_DATE, &date,
                           -1);

        assert(message && number && date);

        const GCharPtr messageFree(message);
        const GCharPtr numberFree(number);
        const GCharPtr dateFree(date);

        GtkWidget* const infoDialog = gtk_message_dialog_new(GTK_WINDOW(data),
                                                             GTK_DIALOG_MODAL,
                                                             GTK_MESSAGE_INFO,
                                                             GTK_BUTTONS_CLOSE,
                                                             gettext(" Number : %s\n Date : %s\n Message : \n%s"),
                                                             number,
                                                             date,
                                                             message);
        assert(infoDialog);

        gtk_window_set_title(GTK_WINDOW(infoDialog), gettext("Message"));

        gtk_dialog_run(GTK_DIALOG(infoDialog));
        gtk_widget_destroy(infoDialog);
    }
}

MainWindow::MainWindow(Usb& usb)
    : mainWindow_(MakeMainWindow()),
      connectFileMenu_(0),
      saveFileMenu_(0),
      saveasFileMenu_(0),
      connectToolButton_(0),
      saveToolButton_(0),
      saveasToolButton_(0),
      clearToolButton_(0)
{
    GtkWidget* const mainBox = MakeMainBox(usb);

    gtk_container_add(GTK_CONTAINER(mainWindow_), mainBox);

    gtk_widget_show_all(mainWindow_);
}

GtkWidget* MainWindow::MakeMainWindow()
{
    GtkWidget* const mainWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    assert(mainWindow);

    gtk_window_set_title(GTK_WINDOW(mainWindow), gMainWindowTitle);
    gtk_window_set_default_size(GTK_WINDOW(mainWindow),
                                gMainWindowWidth,
                                gMainWindowHeight);

    GdkGeometry mainWindowHints;
    mainWindowHints.min_width = gMainWindowMinWidth;
    mainWindowHints.min_height = gMainWindowMinHeight;
    gtk_window_set_geometry_hints(GTK_WINDOW(mainWindow),
                                  mainWindow,
                                  &mainWindowHints,
                                  GDK_HINT_MIN_SIZE);

    gtk_window_set_icon_from_file(GTK_WINDOW(mainWindow),
                                  gMainWindowIconFilename,
                                  NULL);

    g_signal_connect(G_OBJECT(mainWindow), "destroy", G_CALLBACK(OnDestory), 0);
    return mainWindow;
}

GtkWidget* MainWindow::MakeFileMenu()
{
    connectFileMenu_ = gtk_image_menu_item_new_from_stock("gtk-connect", 0);
    assert(connectFileMenu_);

    GtkWidget* const separator1FileMenu = gtk_separator_menu_item_new();
    assert(separator1FileMenu);

    saveFileMenu_ = gtk_image_menu_item_new_from_stock("gtk-save", 0);
    assert(saveFileMenu_);

    saveasFileMenu_ = gtk_image_menu_item_new_from_stock("gtk-save-as", 0);
    assert(saveasFileMenu_);

    GtkWidget* const separator2FileMenu = gtk_separator_menu_item_new();
    assert(separator2FileMenu);

    GtkWidget* const quitFileMenu = gtk_image_menu_item_new_from_stock("gtk-quit", 0);
    assert(quitFileMenu);
    g_signal_connect(G_OBJECT(quitFileMenu), "activate", G_CALLBACK(OnDestory), 0);

    GtkWidget* const fileMenu = gtk_menu_new();
    assert(fileMenu);
    gtk_menu_append(fileMenu, connectFileMenu_);
    gtk_menu_append(fileMenu, separator1FileMenu);
    gtk_menu_append(fileMenu, saveFileMenu_);
    gtk_menu_append(fileMenu, saveasFileMenu_);
    gtk_menu_append(fileMenu, separator2FileMenu);
    gtk_menu_append(fileMenu, quitFileMenu);

    return fileMenu;
}

GtkWidget* MainWindow::MakeHelpMenu()
{
    GtkWidget* const aboutHelpMenu = gtk_image_menu_item_new_from_stock("gtk-about", 0);
    assert(aboutHelpMenu);
    g_signal_connect(G_OBJECT(aboutHelpMenu),
                     "activate",
                     G_CALLBACK(OnActivateAbout),
                     mainWindow_);
    GtkWidget* const helpMenu = gtk_menu_new();
    assert(helpMenu);
    gtk_menu_append(helpMenu, aboutHelpMenu);

    return helpMenu;
}

GtkWidget* MainWindow::MakeMenuBar()
{
    GtkWidget* const fileMenu = MakeFileMenu();

    GtkWidget* const fileMenuItem = gtk_menu_item_new_with_mnemonic(gettext("File(_F)"));
    assert(fileMenuItem);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(fileMenuItem), fileMenu);

    GtkWidget* const helpMenu = MakeHelpMenu();

    GtkWidget* const helpMenuItem = gtk_menu_item_new_with_mnemonic(gettext("Help(_H)"));
    assert(helpMenuItem);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(helpMenuItem), helpMenu);

    GtkWidget* const menuBar = gtk_menu_bar_new();
    assert(menuBar);
    gtk_menu_bar_append(menuBar, fileMenuItem);
    gtk_menu_bar_append(menuBar, helpMenuItem);

    return menuBar;
}

GtkWidget* MainWindow::MakeToolBar()
{
    connectToolButton_ = gtk_tool_button_new_from_stock("gtk-connect");
    assert(connectToolButton_);

    saveToolButton_ = gtk_tool_button_new_from_stock("gtk-save");
    assert(saveToolButton_);

    saveasToolButton_ = gtk_tool_button_new_from_stock("gtk-save-as");
    assert(saveasToolButton_);

    clearToolButton_ = gtk_tool_button_new_from_stock("gtk-clear");
    assert(clearToolButton_);

    GtkWidget* const toolBar = gtk_toolbar_new();
    assert(toolBar);

    gtk_toolbar_insert(GTK_TOOLBAR(toolBar), connectToolButton_, -1);
    gtk_toolbar_insert(GTK_TOOLBAR(toolBar), saveToolButton_, -1);
    gtk_toolbar_insert(GTK_TOOLBAR(toolBar), saveasToolButton_, -1);
    gtk_toolbar_insert(GTK_TOOLBAR(toolBar), clearToolButton_, -1);

    return toolBar;
}

GtkWidget* MainWindow::MakeSMSTreeView()
{
    GtkWidget* const smsTreeView = gtk_tree_view_new();
    assert(smsTreeView);

    GtkCellRenderer* const smsMsgRenderer = gtk_cell_renderer_text_new();
    assert(smsMsgRenderer);
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(smsTreeView),
                                                -1,
                                                gettext("Message"),
                                                smsMsgRenderer,
                                                "text",
                                                SMS_COL_MESSAGE,
                                                NULL);

    GtkCellRenderer* const smsNumRenderer = gtk_cell_renderer_text_new();
    assert(smsNumRenderer);
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(smsTreeView),
                                                -1,
                                                gettext("Number"),
                                                smsNumRenderer,
                                                "text",
                                                SMS_COL_NUMBER,
                                                NULL);

    GtkCellRenderer* const smsDateRenderer = gtk_cell_renderer_text_new();
    assert(smsDateRenderer);
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(smsTreeView),
                                                -1,
                                                gettext("Date"),
                                                smsDateRenderer,
                                                "text",
                                                SMS_COL_DATE,
                                                NULL);

    GtkListStore* const smsModel = gtk_list_store_new(SMS_NUM_COLS,
                                                      G_TYPE_STRING,
                                                      G_TYPE_STRING,
                                                      G_TYPE_STRING);
    assert(smsModel);
    gtk_tree_view_set_model(GTK_TREE_VIEW(smsTreeView),
                            GTK_TREE_MODEL(smsModel));
    g_object_unref(smsModel);

    g_signal_connect(G_OBJECT(smsTreeView), "row-activated", G_CALLBACK(OnSMSRowActivated), mainWindow_);

    return smsTreeView;
}

GtkWidget* MainWindow::MakePhonebookTreeView()
{
    GtkWidget* const phonebookTreeView = gtk_tree_view_new();
    assert(phonebookTreeView);

    GtkCellRenderer* const phonebookNameRenderer = gtk_cell_renderer_text_new();
    assert(phonebookNameRenderer);
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(phonebookTreeView),
                                                -1,
                                                gettext("Name"),
                                                phonebookNameRenderer,
                                                "text",
                                                PHONE_COL_NAME,
                                                NULL);

    GtkTreeStore* const phonebookModel = gtk_tree_store_new(PHONE_NUM_COLS,
                                                            G_TYPE_STRING);

    assert(phonebookModel);
    gtk_tree_view_set_model(GTK_TREE_VIEW(phonebookTreeView),
                            GTK_TREE_MODEL(phonebookModel));
    g_object_unref(phonebookModel);

    return phonebookTreeView;
}

GtkWidget* MainWindow::MakePhotoTreeView()
{
    GtkWidget* const photoTreeView = gtk_tree_view_new();
    assert(photoTreeView);

    GtkCellRenderer* const photoNameRenderer = gtk_cell_renderer_text_new();
    assert(photoNameRenderer);
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(photoTreeView),
                                                -1,
                                                gettext("Name"),
                                                photoNameRenderer,
                                                "text",
                                                FILE_COL_NAME,
                                                NULL);

    GtkCellRenderer* const photoSizeRenderer = gtk_cell_renderer_text_new();
    assert(photoSizeRenderer);
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(photoTreeView),
                                                -1,
                                                gettext("Size"),
                                                photoSizeRenderer,
                                                "text",
                                                FILE_COL_SIZE,
                                                NULL);

    GtkListStore* const photoModel = gtk_list_store_new(FILE_NUM_COLS,
                                                        G_TYPE_STRING,
                                                        G_TYPE_STRING);
    assert(photoModel);
    gtk_tree_view_set_model(GTK_TREE_VIEW(photoTreeView),
                            GTK_TREE_MODEL(photoModel));
    g_object_unref(photoModel);

    return photoTreeView;
}

GtkWidget* MainWindow::MakeVideoTreeView()
{
    GtkWidget* const videoTreeView = gtk_tree_view_new();
    assert(videoTreeView);

    GtkCellRenderer* const videoNameRenderer = gtk_cell_renderer_text_new();
    assert(videoNameRenderer);
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(videoTreeView),
                                                -1,
                                                gettext("Name"),
                                                videoNameRenderer,
                                                "text",
                                                FILE_COL_NAME,
                                                NULL);
    GtkCellRenderer* const videoSizeRenderer = gtk_cell_renderer_text_new();
    assert(videoSizeRenderer);
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(videoTreeView),
                                                -1,
                                                gettext("Size"),
                                                videoSizeRenderer,
                                                "text",
                                                FILE_COL_SIZE,
                                                NULL);
    GtkListStore* const videoModel = gtk_list_store_new(FILE_NUM_COLS,
                                                        G_TYPE_STRING,
                                                        G_TYPE_STRING);
    assert(videoModel);
    gtk_tree_view_set_model(GTK_TREE_VIEW(videoTreeView),
                            GTK_TREE_MODEL(videoModel));
    g_object_unref(videoModel);

    return videoTreeView;
}

GtkWidget* MainWindow::MakeNotebook()
{
    GtkWidget* const smsTreeView = MakeSMSTreeView();
    GtkWidget* const smsTabLabel = gtk_label_new(gettext("SMS"));
    assert(smsTabLabel);

    GtkWidget* const phonebookTreeView = MakePhonebookTreeView();
    GtkWidget* const phonebookTabLabel = gtk_label_new(gettext("Phonebook"));
    assert(phonebookTabLabel);

    GtkWidget* const photoTreeView = MakePhotoTreeView();
    GtkWidget* const photoTabLabel = gtk_label_new(gettext("Photo"));
    assert(photoTabLabel);

    GtkWidget* const videoTreeView = MakeVideoTreeView();
    GtkWidget* const videoTabLabel = gtk_label_new(gettext("Video"));
    assert(videoTabLabel);

    GtkWidget* const notebook = gtk_notebook_new();
    assert(notebook);

    GtkWidget* const smsScrolledWindow = gtk_scrolled_window_new(NULL, NULL);
    assert(smsScrolledWindow);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(smsScrolledWindow),
                                   GTK_POLICY_AUTOMATIC,
                                   GTK_POLICY_AUTOMATIC);
    gtk_container_add(GTK_CONTAINER(smsScrolledWindow), smsTreeView);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
                             smsScrolledWindow,
                             smsTabLabel);


    GtkWidget* const phonebookScrolledWindow = gtk_scrolled_window_new(NULL, NULL);
    assert(phonebookScrolledWindow);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(phonebookScrolledWindow),
                                   GTK_POLICY_AUTOMATIC,
                                   GTK_POLICY_AUTOMATIC);
    gtk_container_add(GTK_CONTAINER(phonebookScrolledWindow),
                      phonebookTreeView);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
                             phonebookScrolledWindow,
                             phonebookTabLabel);


    GtkWidget* const photoScrolledWindow = gtk_scrolled_window_new(NULL, NULL);
    assert(photoScrolledWindow);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(photoScrolledWindow),
                                   GTK_POLICY_AUTOMATIC,
                                   GTK_POLICY_AUTOMATIC);
    gtk_container_add(GTK_CONTAINER(photoScrolledWindow),
                      photoTreeView);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
                             photoScrolledWindow,
                             photoTabLabel);


    GtkWidget* const videoScrolledWindow = gtk_scrolled_window_new(NULL, NULL);
    assert(videoScrolledWindow);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(videoScrolledWindow),
                                   GTK_POLICY_AUTOMATIC,
                                   GTK_POLICY_AUTOMATIC);
    gtk_container_add(GTK_CONTAINER(videoScrolledWindow),
                      videoTreeView);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
                             videoScrolledWindow,
                             videoTabLabel);

    g_signal_connect(G_OBJECT(notebook),
                     "switch-page",
                     G_CALLBACK(OnSwitchPage),
                     sMenu_.get());

    return notebook;
}

void MainWindow::MenuSignalConnect(Usb& usb, GtkNotebook& notebook)
{
    tArg_ = std::auto_ptr<ConnectThreadArg>(new ConnectThreadArg(usb, notebook, *sMenu_));

    g_signal_connect(G_OBJECT(connectFileMenu_),
                     "activate",
                     G_CALLBACK(OnActivateConnect),
                     tArg_.get());
    g_signal_connect(G_OBJECT(connectToolButton_),
                     "clicked",
                     G_CALLBACK(OnActivateConnect),
                     tArg_.get());

    sArg_ = std::auto_ptr<SaveWidgetArg>(new SaveWidgetArg(usb, notebook));

    g_signal_connect(G_OBJECT(saveFileMenu_),
                     "activate",
                     G_CALLBACK(OnActivateSave),
                     sArg_.get());
    g_signal_connect(G_OBJECT(saveToolButton_),
                     "clicked",
                     G_CALLBACK(OnActivateSave),
                     sArg_.get());

    g_signal_connect(G_OBJECT(saveasFileMenu_),
                     "activate",
                     G_CALLBACK(OnActivateSaveAs),
                     sArg_.get());
    g_signal_connect(G_OBJECT(saveasToolButton_),
                     "clicked",
                     G_CALLBACK(OnActivateSaveAs),
                     sArg_.get());

    cArg_ = std::auto_ptr<ClearWidgetArg>(new ClearWidgetArg(notebook, *sMenu_));

    g_signal_connect(G_OBJECT(clearToolButton_),
                     "clicked",
                     G_CALLBACK(OnClickedClear),
                     cArg_.get());
}

GtkWidget* MainWindow::MakeMainBox(Usb& usb)
{
    GtkWidget* const mainMenuBar = MakeMenuBar();

    GtkWidget* const mainToolBar = MakeToolBar();

    sMenu_ = std::auto_ptr<SensitiveMenu>(new SensitiveMenu(*saveFileMenu_,
                                                            *saveasFileMenu_,
                                                            *(GTK_WIDGET(saveToolButton_)),
                                                            *(GTK_WIDGET(saveasToolButton_)),
                                                            *(GTK_WIDGET(clearToolButton_))));

    GtkWidget* const notebook = MakeNotebook();

    GtkWidget* const statusBar = gtk_statusbar_new();
    assert(statusBar);

    GtkWidget* const mainBox = gtk_vbox_new(0, 0);
    assert(mainBox);

    gtk_box_pack_start(GTK_BOX(mainBox), mainMenuBar, 0, 1, 0);
    gtk_box_pack_start(GTK_BOX(mainBox), mainToolBar, 0, 1, 0);
    gtk_box_pack_start(GTK_BOX(mainBox), notebook, 1, 1, 0);
    gtk_box_pack_start(GTK_BOX(mainBox), statusBar, 0, 1, 0);

    MenuSignalConnect(usb, *(GTK_NOTEBOOK(notebook)));

    return mainBox;
}

int main(int argc, char* argv[])
{
#ifdef ENABLE_NLS
    setlocale(LC_ALL, "");
    bindtextdomain(GETTEXT_PACKAGE, LOCALE_DIR);
    bind_textdomain_codeset(GETTEXT_PACKAGE, gUtf8CodesetStr);
    textdomain(GETTEXT_PACKAGE);
#endif

    g_thread_init(NULL);
    gdk_threads_init();

    gtk_init(&argc, &argv);

    std::auto_ptr<Usb> thisUsb;

    try {

        thisUsb = std::auto_ptr<Usb>(new Usb);

    } catch(std::exception& e) {

        printf("%s\n", e.what());
        
        GtkWidget* const errDialog = gtk_message_dialog_new(gMainWindow,
                                                           GTK_DIALOG_MODAL,
                                                           GTK_MESSAGE_ERROR,
                                                           GTK_BUTTONS_CLOSE,
                                                           e.what());
        assert(errDialog);
        
        gtk_dialog_run(GTK_DIALOG(errDialog));

        gtk_widget_destroy(errDialog);

        return 1;
    }

    assert(thisUsb.get());

    MainWindow mainWindow(*thisUsb);

    gMainWindow = mainWindow.Get();
    assert(gMainWindow);

    gdk_threads_enter();
    gtk_main();
    gdk_threads_leave();

    return 0;
}
