/*****************************************************************************
 *
 *  gkv5500 - gtk2 frontend of kv5500 utility.
 *
 * Copyright (c) 2006-2007 Hyo geol, Lee
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 *   3. The name of the author may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************/
#ifndef GKV5500_H
#define GKV5500_H

#include <memory>
#include <string>
#include <vector>

#include "auto_array.hpp"

#include <config.h>

namespace KV5500 {

    enum { SMS_PAGE = 0, PHONEBOOK_PAGE, PHOTO_PAGE, VIDEO_PAGE };

    enum { SMS_COL_MESSAGE = 0, SMS_COL_NUMBER, SMS_COL_DATE, SMS_NUM_COLS };
    enum { PHONE_COL_NAME = 0, PHONE_NUM_COLS };

    enum { FILE_COL_NAME = 0, FILE_COL_SIZE, FILE_NUM_COLS };

    const size_t gDateStrSize = 20;
    const double gDateStrGrowFactor = 1.5;

    const char* gUtf8CodesetStr = "UTF-8";
    const char* gKV5500CodesetStr = "EUC-KR";

    const int gTransferDialogWidth = 261;
    const int gTransferDialogHeight = 261;

    const char* gMainWindowTitle = "gkv5500 " VERSION;
    const int gMainWindowWidth = 290;
    const int gMainWindowHeight = 468;
    const int gMainWindowMinWidth = 290;
    const int gMainWindowMinHeight = 300;

    const char* gMainWindowIconFilename = PIXMAPS_DIR "gkv5500.png";
    extern GtkWindow* gMainWindow;

    class SensitiveMenu {
    public:
        SensitiveMenu(GtkWidget& saveMenu,
                      GtkWidget& saveasMenu,
                      GtkWidget& saveToolbar,
                      GtkWidget& saveasToolbar,
                      GtkWidget& clearToolbar);

        void On();
        void Off();
        void Enable(bool flag);

    private:
        GtkWidget& save_;
        GtkWidget& saveas_;
        GtkWidget& saveTool_;
        GtkWidget& saveasTool_;
        GtkWidget& clearTool_;

        SensitiveMenu& operator=(const SensitiveMenu&);
    };

    struct Usb;
    struct ConnectThreadArg {
        ConnectThreadArg(Usb& usb, GtkNotebook& notebook, SensitiveMenu& sMenu);

        Usb& usb_;
        GtkDialog* dialog_;
        GtkLabel* msgLabel_;
        GtkWidget* closeBtn_;
        GtkNotebook& notebook_;
        SensitiveMenu& sMenu_;

    private:
        ConnectThreadArg(const ConnectThreadArg&);
        ConnectThreadArg& operator=(const ConnectThreadArg);
    };

    struct TransferThreadArg {
        TransferThreadArg(Usb& usb, GtkLabel& msgLabel, GtkWidget& progressBar, GtkWidget& closeBtn, unsigned int fileNum, const std::string& filename);

        Usb& usb_;
        GtkLabel& msgLabel_;
        GtkWidget& progressBar_;
        GtkWidget& closeBtn_;

        const unsigned int fileNum_;
        const std::string filename_;

    private:
        TransferThreadArg(const TransferThreadArg&);
        TransferThreadArg& operator=(const TransferThreadArg);
    };

    struct SaveWidgetArg {
        SaveWidgetArg(Usb& usb, GtkNotebook& notebook);

        Usb& usb_;
        GtkNotebook& notebook_;

    private:
        SaveWidgetArg(const SaveWidgetArg&);
        SaveWidgetArg& operator=(const SaveWidgetArg);
    };

    struct ClearWidgetArg {
        ClearWidgetArg(GtkNotebook& notebook, SensitiveMenu& sMenu);

        GtkNotebook& notebook_;
        SensitiveMenu& sMenu_;

    private:
        ClearWidgetArg(const ClearWidgetArg&);
        ClearWidgetArg& operator=(const ClearWidgetArg);
    };

    class MainWindow {
    public:
        MainWindow(Usb& usb);

        GtkWindow* Get();

    private:
        GtkWidget* mainWindow_;

        GtkWidget* connectFileMenu_;
        GtkWidget* saveFileMenu_;
        GtkWidget* saveasFileMenu_;

        GtkToolItem* connectToolButton_;
        GtkToolItem* saveToolButton_;
        GtkToolItem* saveasToolButton_;
        GtkToolItem* clearToolButton_;

        std::auto_ptr<SensitiveMenu> sMenu_;
        std::auto_ptr<ConnectThreadArg> tArg_;
        std::auto_ptr<SaveWidgetArg> sArg_;
        std::auto_ptr<ClearWidgetArg> cArg_;

        GtkWidget* MakeMainWindow();

        GtkWidget* MakeFileMenu();
        GtkWidget* MakeHelpMenu();
        GtkWidget* MakeMenuBar();

        GtkWidget* MakeToolBar();

        GtkWidget* MakeSMSTreeView();
        GtkWidget* MakePhonebookTreeView();
        GtkWidget* MakePhotoTreeView();
        GtkWidget* MakeVideoTreeView();

        GtkWidget* MakeNotebook();

        void MenuSignalConnect(Usb& usb, GtkNotebook& notebook);

        GtkWidget* MakeMainBox(Usb& usb);

        MainWindow(const MainWindow&);
        MainWindow& operator=(const MainWindow&);
    };

    std::ostream& PrintSeparator(std::ostream& os, unsigned int current, unsigned int total);

    void MakeConnectDialog(ConnectThreadArg& tArg, const gchar* title, const gchar* defaultLabel);
    GtkWidget* MakeSaveAsDialog(const char* const defaultFilename);

    const std::string GetFilesizeStr(unsigned int size);
    const int GetFileNumber(GtkNotebook& notebook, gint page);

    typedef HG::AutoArray<char>::Type CharArrayPtr;
    CharArrayPtr DateStrNow();

    const std::string GetDateFilename(const char* prefix, const char* postfix);

    GtkTreeView* GetTreeView(GtkNotebook& notebook, gint page);
    GtkTreeModel* GetTreeModel(GtkNotebook& notebook, gint page);

    static void OnDestory(GtkWidget*, gpointer);

    class SMS;
    void RegistSMS(const std::vector<SMS>& sms, GtkListStore& listStore);

    class PhonebookData;
    void RegistPhonebook(const std::vector<PhonebookData>& phonebook, GtkTreeStore& treeStore);

    class FileInfo;
    void RegistFileInfo(const std::vector<FileInfo>& fileInfo, GtkListStore& listStore);

    int SMSScan(Usb& usb, GtkNotebook& notebook);
    int PhonebookScan(Usb& usb, GtkNotebook& notebook);
    int PhotoInfoScan(Usb& usb, GtkNotebook& notebook);
    int VideoInfoScan(Usb& usb, GtkNotebook& notebook);

    void* GetSMSThread(void* arg);
    void* GetPhonebookThread(void* arg);
    void* GetPhotoListThread(void* arg);
    void* GetVideoListThread(void* arg);

    void ConnectSMS(ConnectThreadArg& tArg);
    void ConnectPhonebook(ConnectThreadArg& tArg);
    void ConnectPhoto(ConnectThreadArg& tArg);
    void ConnectVideo(ConnectThreadArg& tArg);
    static void OnActivateConnect(GtkMenuItem*, gpointer data);

    void ShowSavedDialog(const char* filename);
    void ShowSaveErrDialog(const char* filename);
    GtkWidget* MakeTransferDialog(GtkLabel*& label, GtkWidget*& progressBar, GtkWidget*& closeBtn);

    void SaveSMS(GtkNotebook& notebook, const char* filename);
    void SavePhonebook(GtkNotebook& notebook, const char* filename);
    void SavePhoto(Usb& usb, unsigned int fileNum, const std::string& filename);
    void SaveVideo(Usb& usb, unsigned int fileNum, const std::string& filename);

    void* TransferPhotoThread(void* arg);
    void* TransferVideoThread(void* arg);

    static void OnActivateSave(GtkMenuItem*, gpointer data);
    static void OnActivateSaveAs(GtkMenuItem*, gpointer data);
    static void OnClickedClear(GtkMenuItem*, gpointer data);
    static void OnActivateAbout(GtkMenuItem*, gpointer data);
    static void OnSwitchPage(GtkNotebook*, GtkNotebookPage*, guint, gpointer);
    static void OnSMSRowActivated(GtkTreeView*, GtkTreePath*, GtkTreeViewColumn*, gpointer);
}

/// inline functions
using namespace KV5500;

inline SensitiveMenu::SensitiveMenu(GtkWidget& saveMenu,
                                    GtkWidget& saveasMenu,
                                    GtkWidget& saveToolbar,
                                    GtkWidget& saveasToolbar,
                                    GtkWidget& clearToolbar)
    : save_(saveMenu),
      saveas_(saveasMenu),
      saveTool_(saveToolbar),
      saveasTool_(saveasToolbar),
      clearTool_(clearToolbar)
{}

inline void SensitiveMenu::Enable(bool flag)
{
    if(flag)
        On();
    else
        Off();
}

inline ConnectThreadArg::ConnectThreadArg(Usb& usb, GtkNotebook& notebook, SensitiveMenu& sMenu)
    : usb_(usb),
      notebook_(notebook),
      sMenu_(sMenu)
{}

inline TransferThreadArg::TransferThreadArg(Usb& usb, GtkLabel& msgLabel, GtkWidget& progressBar, GtkWidget& closeBtn, unsigned int fileNum, const std::string& filename)
    : usb_(usb),
      msgLabel_(msgLabel),
      progressBar_(progressBar),
      closeBtn_(closeBtn),
      fileNum_(fileNum),
      filename_(filename)
{}

inline SaveWidgetArg::SaveWidgetArg(Usb& usb, GtkNotebook& notebook)
    : usb_(usb),
      notebook_(notebook)
{}

inline ClearWidgetArg::ClearWidgetArg(GtkNotebook& notebook, SensitiveMenu& sMenu)
    : notebook_(notebook),
      sMenu_(sMenu)
{}

inline GtkWindow* MainWindow::Get()
{
    return GTK_WINDOW(mainWindow_);
}

#endif
