/*****************************************************************************
 *
 *  kv5500 - utility library for kv5500 like phone.
 *
 * Copyright (c) 2006-2007 Hyo geol, Lee
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 *   3. The name of the author may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************/
#ifndef KV5500_H
#define KV5500_H

#include <string>
#include <exception>

#include <assert.h>

#include <usb.h>

namespace KV5500 {

    /// @brief Exception while no bus found.
    class UsbNoBus : public std::exception {
    public:
        const char* what() const throw();
    };

    /// @brief Exception while no device found.
    class UsbNoDev : public std::exception {
    public:
        const char* what() const throw();
    };

    /// @brief Exception while usb_open
    class UsbOpen : public std::exception {
    public:
        const char* what() const throw();
    };

    /// @brief Simple usb library wrapper for kv5500.
    class Usb {
    public:
        /// @param test Test bus and device changes.
        Usb();

        template <typename T>
        bool Scan(T& obj);

        bool Find();

    private:
        /// don't copy this object.
        Usb(const Usb&);
        Usb& operator=(const Usb&);
    };

    /// @brief Simple usb library wrapper for kv5500. Device section.
    class UsbDev {
    public:
        UsbDev(struct usb_device* dev);

        ~UsbDev();

        int BulkRead(char* bytes, int size, int timeout);
        int BulkWrite(char* bytes, int size, int timeout);

        int Reset();

    private:
        usb_dev_handle* const udev_;

        /// don't copy this object.
        UsbDev(const UsbDev&);
        UsbDev& operator=(const UsbDev&);
    };

    bool DescCheck(struct usb_device* dev);
    bool WakeUp(UsbDev& udev);
    bool Shutdown(UsbDev& udev);

    /// @brief Find endCh to maxEndPoint in buf, return copy string.
    std::string GetString(const char* const buf, const size_t maxEndPoint, const unsigned int offset, const size_t maxStrSize, const char endCh = 0x00);
}

/// inline functions.
using namespace KV5500;

template <typename T>
bool Usb::Scan(T& obj)
{
    bool rtn = false;

    for (usb_bus* bus = usb_busses; bus; bus = bus->next) {

        if (bus->root_dev)
            rtn = rtn || obj.Scan(bus->root_dev);
        else {

            for (struct usb_device* dev = bus->devices; dev; dev = dev->next)
                rtn = rtn || obj.Scan(dev);
        }
    }

    return rtn;
}

inline bool Usb::Find()
{
    return usb_find_busses() >= 0 && usb_find_devices >= 0;
}

inline UsbDev::UsbDev(struct usb_device* dev)
    : udev_(usb_open(dev))
{
    if(!udev_)
        throw UsbOpen();
}

inline UsbDev::~UsbDev()
{
    usb_close(udev_);
}

inline int UsbDev::BulkRead(char* bytes, int size, int timeout)
{
    return usb_bulk_read(udev_, 0x84, bytes, size, timeout);
}

inline int UsbDev::BulkWrite(char* bytes, int size, int timeout)
{
    return usb_bulk_write(udev_, 0x04, bytes, size, timeout);
}

inline int UsbDev::Reset()
{
    return usb_reset(udev_);
}

/// vendor 1004 and product 6000, and this is not correct test.
inline bool KV5500::DescCheck(struct usb_device* dev)
{
    assert(dev);

    return dev->descriptor.idVendor == 0x1004 &&
        dev->descriptor.idProduct == 0x6000;
}

#endif
