/*****************************************************************************
 *
 *  kv5500 - utility library for kv5500 like phone.
 *
 * Copyright (c) 2006-2007 Hyo geol, Lee
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 *   3. The name of the author may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************/
#ifndef KV5500_FILE_H
#define KV5500_FILE_H

#include <string>
#include <vector>
#include <sstream>
#include <fstream>
#include <exception>

#include <cstdio>
#include <cstring>

#include <time.h>
#include <assert.h>

#include "kv5500.hpp"

namespace KV5500 {

    class UsbDev;

    class FileInfo {
    public:
        FileInfo(const std::string& name, unsigned int size);

        const std::string& GetName() const;
        unsigned int GetSize() const;

        void SetName(const char* str);
        void SetSize(unsigned int size);

    private:
        std::string name_;
        unsigned int fileSize_;
    };

    void PrintFileInfo(const FileInfo& fi);

    template <typename T>
    class FileInfoScanner {
    public:
        FileInfoScanner();

        void GetInfo(std::vector<FileInfo>& vec) const;
        unsigned int GetInfoNum() const;

        bool Scan(struct usb_device* dev);

    private:
        char numReqTempl_[5];
        const size_t numReqSize_;
        
        char nameReqTempl_[6];
        const size_t nameReqSize_;

        char sizeReqTempl_[8];
        const size_t sizeReqSize_;

        std::vector<FileInfo> fileInfoContainer_;

        bool Request(UsbDev& udev);

        FileInfoScanner(const FileInfoScanner&);
        FileInfoScanner& operator=(const FileInfoScanner&);
    };

    void PrintFileInfoVec(const std::vector<FileInfo>& v);

    class FileGetNum : public std::exception {
        const char* what() const throw();
    };

    template <typename T, typename NameGenerator, typename Display>
    class GetFileScanner {
    public:
        GetFileScanner(unsigned int num, NameGenerator& nameGenerator, Display& display);

        bool Scan(struct ::usb_device* dev);

    private:
        char numReqTempl_[5];
        const size_t numReqSize_;

        char sizeReqTempl_[8];
        const size_t sizeReqSize_;

        char prepareReqTempl_[5];
        const size_t prepareReqSize_;

        char transReqTempl_[9];
        const size_t transReqSize_;

        const unsigned int num_;

        NameGenerator& nameGenerator_;
        Display& display_;

        bool CheckFileNum(UsbDev& udev);

        bool SendPrepareTransfer(UsbDev& udev);

        bool PutFile(UsbDev& udev);

        GetFileScanner(const GetFileScanner&);
        GetFileScanner& operator=(const GetFileScanner&);
    };

    struct ConsoleDisplay {
        void PrintError(const std::string& msg);
        void PrintError(const char* msg);
        void Print(const std::string& msg);
        void PrintProcess(unsigned int complete, unsigned int total);
        void PrintEnd(const std::string& msg);
    };

    /// @brief Get 4bytes reverse endian.
    unsigned int GetRevEndian(char* buf);

    /// @brief Recv file name from udev to name.
    int RecvName(UsbDev& udev, std::string& name);

    /// @brief Recv file size from udev to size.
    int RecvSize(UsbDev& udev, unsigned int& size);

    bool GetTotalNum(UsbDev& udev, char* const req, size_t reqSize, unsigned int& num);

    /// @brief Recv file number from udev to size.
    bool RecvNum(UsbDev& udev, unsigned int& num);

    void PrintTransferProcess(unsigned int complete, unsigned int total);

    /// @brief Check buf and buf + 1 is 7d 5e or 7d 5d.
    bool HasSubstStr(const char* const buf, const unsigned int pos);

    /// @brief Count substituted 12 bytes from begin.
    unsigned int GetPrefixPos(const char* const buf);
    /// @brief Count substituted 3 bytes from end.
    unsigned int GetPostfixPos(const char* const buf, const unsigned int bufSize);
}

/// inline functions.
using namespace KV5500;

inline FileInfo::FileInfo(const std::string& name, unsigned int size)
    : name_(name),
      fileSize_(size)
{}

inline const std::string& FileInfo::GetName() const
{
    return name_;
}

inline unsigned int FileInfo::GetSize() const
{
    return fileSize_;
}

inline void FileInfo::SetName(const char* str)
{
    assert(str);

    name_ = str;
}

inline void FileInfo::SetSize(unsigned int size)
{
    fileSize_ = size;
}

inline void KV5500::PrintFileInfo(const FileInfo& fi)
{
    printf(" Name  : %s\n", fi.GetName().c_str());
    printf(" Size  : %.2f MBytes\n", fi.GetSize() / 1024.0 / 1024.0);
}

template <typename T>
inline FileInfoScanner<T>::FileInfoScanner()
    : numReqSize_(5),
      nameReqSize_(6),
      sizeReqSize_(8)
{
    memcpy(numReqTempl_, T::numReqTempl_, numReqSize_);
    memcpy(nameReqTempl_, T::nameReqTempl_, nameReqSize_);
    memcpy(sizeReqTempl_, T::sizeReqTempl_, sizeReqSize_);
}

template <typename T>
inline void FileInfoScanner<T>::GetInfo(std::vector<FileInfo>& vec) const
{
    vec = fileInfoContainer_;
}

template <typename T>
inline unsigned int FileInfoScanner<T>::GetInfoNum() const
{
    return fileInfoContainer_.size();
}

template <typename T>
bool FileInfoScanner<T>::Scan(struct usb_device* dev)
{
    assert(dev);

    bool rtn = false;

    try {

        UsbDev udev(dev);

        /// vendor 1004 and product 6000, and this is not correct test.
        if(DescCheck(dev)) {

            if(!WakeUp(udev))
                throw std::exception();

            if(!Request(udev))
                throw std::exception();

            if(!Shutdown(udev))
                throw std::exception();

            rtn = true;
        }

    } catch(std::exception& e) {

        printf("%s\n", e.what());
    }

    for (int i = 0; i < dev->num_children; i++)
        rtn = rtn || Scan(dev->children[i]);

    return rtn;
}

template <typename T>
bool FileInfoScanner<T>::Request(UsbDev& udev)
{
    unsigned int totalNum = 0;
    if(!GetTotalNum(udev, numReqTempl_, numReqSize_, totalNum) || totalNum > 256)
        return false;

    if(totalNum == 0)
        return true;

    std::string name;
    unsigned int fileSize = 0;
    for(unsigned int i = 0; i < totalNum; ++i) {

        nameReqTempl_[3] = i + 1;
        if(udev.BulkWrite(nameReqTempl_, nameReqSize_, 5000) < 0) {

            printf("Got error while send.\n");

            return false;
        }

        if(RecvName(udev, name) < 0)
            return false;

        sizeReqTempl_[3] = i + 1;
        if(udev.BulkWrite(sizeReqTempl_, sizeReqSize_, 5000) < 0) {

            printf("Got error while send.\n");

            return false;
        }

        if(RecvSize(udev, fileSize) < 0)
            return false;

        fileInfoContainer_.push_back(FileInfo(name, fileSize));
    }

    return true;
}

template <typename T, typename N, typename D>
inline GetFileScanner<T, N, D>::GetFileScanner(unsigned int num, N& n, D& d)
    : numReqSize_(5),
      sizeReqSize_(8),
      prepareReqSize_(5),
      transReqSize_(9),
      num_(num),
      nameGenerator_(n),
      display_(d)
{
    if(num == 0 || num > 256)
        throw FileGetNum();

    memcpy(numReqTempl_, T::numReqTempl_, numReqSize_);
    memcpy(sizeReqTempl_, T::sizeReqTempl_, sizeReqSize_);
    memcpy(prepareReqTempl_, T::prepareReqTempl_, prepareReqSize_);
    memcpy(transReqTempl_, T::transReqTempl_, transReqSize_);
}

template <typename T, typename N, typename D>
inline bool GetFileScanner<T, N, D>::Scan(struct usb_device* dev)
{
    assert(dev);

    bool rtn = false;

    try {

        UsbDev udev(dev);

        /// vendor 1004 and product 6000, and this is not correct test.
        if(DescCheck(dev)) {

            if(!WakeUp(udev))
                throw std::exception();

            if(!CheckFileNum(udev))
                throw std::exception();

            if(!PutFile(udev))
                throw std::exception();

            if(!Shutdown(udev))
                throw std::exception();

            rtn = true;
        }

    } catch(std::exception& e) {

        display_.PrintError(e.what());
    }

    for (int i = 0; i < dev->num_children; i++)
        rtn = rtn || Scan(dev->children[i]);

    return true;
}

template <typename T, typename N, typename D>
inline bool GetFileScanner<T, N, D>::CheckFileNum(UsbDev& udev)
{
    unsigned int totalNum = 0;
    GetTotalNum(udev, numReqTempl_, numReqSize_, totalNum);
    if(totalNum > 256 || num_ > totalNum) {

        std::stringstream ss;

        ss << "Total " << totalNum << " files.\n";

        display_.PrintError(ss.str());

        return false;
    }

    return true;
}

template <typename T, typename N, typename D>
inline bool GetFileScanner<T, N, D>::SendPrepareTransfer(UsbDev& udev)
{
    if(udev.BulkWrite(prepareReqTempl_, prepareReqSize_, 5000) < 0) {

        display_.PrintError("Got error while send.\n");

        return false;
    }

    char buf[1000];
    udev.BulkRead(buf, 1000, 5000);

    return true;
}

template <typename T, typename N, typename D>
inline bool GetFileScanner<T, N, D>::PutFile(UsbDev& udev)
{
    const std::string filename = nameGenerator_.GetFilename(num_);

    sizeReqTempl_[3] = num_;
    if(udev.BulkWrite(sizeReqTempl_, sizeReqSize_, 5000) < 0) {

        display_.PrintError("Got error while send.\n");

        return false;
    }

    char buf[3072];
    const int ret = udev.BulkRead(buf, sizeof(buf), 5000);

    /// assume ret must be over 11 bytes.
    if(ret <= 11) {

        display_.PrintError("Got error while recv.\n");

        return false;
    }

    const unsigned int fileSize = GetRevEndian(buf + 8);

    int leftSize = fileSize;
    std::ofstream os(filename.c_str());
    if(!os) {

        display_.PrintError("Cannot open file.\n");

        return false;
    }

    SendPrepareTransfer(udev);

    std::stringstream ss;
    ss << "Writing " << filename << " ......\n";
    display_.Print(ss.str());

    unsigned int i = 0;
    transReqTempl_[3] = num_;
    while(leftSize > 0) {

        display_.PrintProcess(fileSize - leftSize, fileSize);

        transReqTempl_[7] = i;
        if(udev.BulkWrite(transReqTempl_, transReqSize_, 5000) < 0) {

            display_.PrintError("\nGot error while send.\n");

            return false;
        }

        const int ret2 = udev.BulkRead(buf, sizeof(buf), 5000);
        /// assume ret must be over 15 bytes.
        if(ret2 <= 15)
            return false;

        const unsigned int prefix = GetPrefixPos(buf);
        const unsigned int postfix = GetPostfixPos(buf, ret2);

        for(unsigned int j = prefix; (j < ret2 - postfix) && (leftSize > 0); ++j, --leftSize) {
            if(static_cast<unsigned char>(buf[j]) == 0x7d &&
               static_cast<unsigned char>(buf[j+1]) == 0x5d) {

                os << static_cast<char>(0x7d);
                    
                ++j;

            } else if(static_cast<unsigned char>(buf[j]) == 0x7d && 
                      static_cast<unsigned char>(buf[j+1]) == 0x5e) {

                os << static_cast<char>(0x7e);

                ++j;

            } else
                os << buf[j];
        }

        if(leftSize > 0 && i == 256)
            i = 0;
        else
            ++i;
    }

    std::stringstream endss;

    if(leftSize == 0) {

        endss << "\rDone. " << fileSize << " Bytes transferred.\n";

        display_.PrintEnd(endss.str());

    } else {

        endss << "\rError. " << leftSize << " left.\n";

        display_.PrintEnd(endss.str());
    }

    return true;
}

inline void ConsoleDisplay::PrintError(const std::string& msg)
{
    printf("%s\n", msg.c_str());
}

inline void ConsoleDisplay::PrintError(const char* msg)
{
    printf("%s\n", msg);
}

inline void ConsoleDisplay::Print(const std::string& msg)
{
    printf("%s\n", msg.c_str());
}

inline void ConsoleDisplay::PrintProcess(unsigned int complete, unsigned int total)
{
    PrintTransferProcess(complete, total);
}

inline void ConsoleDisplay::PrintEnd(const std::string& msg)
{
    printf("%s\n", msg.c_str());
}

inline unsigned int KV5500::GetRevEndian(char* buf)
{
    assert(buf);

    return static_cast<unsigned char>(buf[0]) + 
        (static_cast<unsigned char>(buf[1]) << 8) + 
        (static_cast<unsigned char>(buf[2]) << 16) +
        (static_cast<unsigned char>(buf[3]) << 24);
}

inline void KV5500::PrintTransferProcess(unsigned int complete, unsigned int total)
{
    assert(complete <= total);

    printf("\r%d / %d (%.2f%%)", complete, total, 100.0 * complete / total);
}

inline bool KV5500::HasSubstStr(const char* const buf, const unsigned int pos)
{
    assert(buf);

    return static_cast<unsigned char>(buf[pos]) == 0x7d &&
        (static_cast<unsigned char>(buf[pos + 1]) == 0x5e ||
         static_cast<unsigned char>(buf[pos + 1]) == 0x5d);
}

inline unsigned int KV5500::GetPostfixPos(const char* const buf, const unsigned int bufSize)
{
    assert(buf && bufSize > 4);

    return HasSubstStr(buf, bufSize - 4) || HasSubstStr(buf, bufSize -3) ? 4 : 3;
}

#endif
