/*****************************************************************************
 *
 *  smartptr - Simple smart point library.
 *
 *             DestPtr - Destructive smart pointer.
 *                       Origin from auto_ptr in MEC++ by Scott Meyers.
 *
 * Copyright (c) 2007 Hyo geol, Lee
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 *   3. The name of the author may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************/
// 	$Id: smartptr.hpp,v 1.2 2007/01/07 18:08:15 ez8 Exp $	
#ifndef SMARTPTR_H
#define SMARTPTR_H

namespace HG {

    /// @brief Helper struct to converse temporary object.
    template <typename T>
    struct DestPtrRef {

        explicit DestPtrRef(T* p);

        T* point_;
    };

    /// @brief Destructive smart point.
    template <typename T, template <class> class Destructor>
    class DestPtr {
    public:
        /// @brief Constructor.
        explicit DestPtr(T* p = 0);

        /// @brief Constructor from other object.
        DestPtr(DestPtr& rhs);

        /// @brief Constructor from other object.
        template <typename U>
        DestPtr(DestPtr<U, Destructor>& rhs);

        /// @brief Constructor from auto_array_ref. cf) temporary object.
        DestPtr(DestPtrRef<T> ref);

        /// @brief Destructor.
        ~DestPtr();

        /// @brief Copy constructor from other object.
        DestPtr& operator=(DestPtr& rhs);

        /// @brief Copy constructor from other object.
        template <typename U>
        DestPtr& operator=(DestPtr<U, Destructor>& rhs);

        /// @brief Copy constructor from auto_array_ref. cf) temporary object.
        DestPtr& operator=(DestPtrRef<T> ref);

        /// @brief Conversion operator to auto_array_ref.
        template <typename U>
        operator DestPtrRef<U>();

        T& operator*() const;
        T* operator->() const;

        /// @brief Get own raw point.
        T* Get() const;

        /// @brief Release own point and return own point.
        T* Release();

        /// @brief Reset to other point. 
        ///        Delete own point and set to other point.
        void Reset(T* p = 0);

    private:
        T* point_;
    };
}

/// inline functions.
using namespace HG;

template <typename T>
inline DestPtrRef<T>::DestPtrRef(T* p)
    : point_(p)
{}

template <typename T, template <class> class Destructor>
inline DestPtr<T, Destructor>::DestPtr(T* p)
    : point_(p)
{}

template <typename T, template <class> class Destructor>
inline DestPtr<T, Destructor>::DestPtr(DestPtr& rhs)
    : point_(rhs.Release())
{}

template <typename T, template <class> class Destructor>
template <typename U>
inline DestPtr<T, Destructor>::DestPtr(DestPtr<U, Destructor>& rhs)
    : point_(rhs.release())
{}

template <typename T, template <class> class Destructor>
inline DestPtr<T, Destructor>::DestPtr(DestPtrRef<T> ref)
    : point_(ref.point_)
{}

template <typename T, template <class> class Destructor>
inline DestPtr<T, Destructor>::~DestPtr()
{
    Destructor<T>::Destroy(point_);
}

template <typename T, template <class> class Destructor>
inline DestPtr<T, Destructor>& DestPtr<T, Destructor>::operator=(DestPtr& rhs)
{
    if(this != &rhs)
        Reset(rhs.Release());

    return *this;
}

template <typename T, template <class> class Destructor>
template <typename U>
inline DestPtr<T, Destructor>& DestPtr<T, Destructor>::operator=(DestPtr<U, Destructor>& rhs)
{
    if(this != &rhs)
        Reset(rhs.Release());

    return *this;
}

template <typename T, template <class> class Destructor>
inline DestPtr<T, Destructor>& DestPtr<T, Destructor>::operator=(DestPtrRef<T> ref)
{
    Reset(ref.point_);

    return *this;
}

template <typename T, template <class> class Destructor>
template <typename U>
inline DestPtr<T, Destructor>::operator DestPtrRef<U>()
{
    return DestPtrRef<U>(this->Release());
}

template <typename T, template <class> class Destructor>
inline T& DestPtr<T, Destructor>::operator*() const
{
    return *point_;
}

template <typename T, template <class> class Destructor>
inline T* DestPtr<T, Destructor>::operator->() const
{
    return point_;
}

template <typename T, template <class> class Destructor>
inline T* DestPtr<T, Destructor>::Get() const
{
    return point_;
}

template <typename T, template <class> class Destructor>
inline T* DestPtr<T, Destructor>::Release()
{
    T* const point = point_;
    point_ = 0;

    return point;
}

template <typename T, template <class> class Destructor>
inline void DestPtr<T, Destructor>::Reset(T* p)
{
    if(point_ != p) {
        Destructor<T>::Destroy(point_);
        point_ = p;
    }
}

#endif
